@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/child.c,v 1.28 2004/06/15 12:57:55 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/child.c,v 1.29 2004/06/30 08:01:52 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -391,7 +391,12 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd)
 	afd = accept(fd, &addr, &addrlen);
 	if (afd < 0)
 	{
-		pool_error("accept() failed. reason: %s", strerror(errno));
+		/*
+		 * "Resource temporarily unavailable" (EAGAIN or EWOULDBLOCK)
+		 * can be silently ignored.
+		 */
+		if (errno != EAGAIN && errno != EWOULDBLOCK)
+			pool_error("accept() failed. reason: %s", strerror(errno));
 		return NULL;
 	}
 #ifdef ACCEPT_PERFORMANCE
