@@ -1,7 +1,7 @@
 /* -*-pgsql-c-*- */
 /*
  *
- * $Header: /home/t-ishii/repository/pgpool/pool.h,v 1.16 2004/05/11 09:27:14 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/pool.h,v 1.19 2004/06/15 12:57:55 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -59,13 +59,54 @@
 #define STRICT_MODE(s) (strncasecmp((s), STRICT_MODE_STR, strlen(STRICT_MODE_STR)) == 0)
 
 typedef enum {
-    POOL_CONTINUE,
+    POOL_CONTINUE = 0,
 	POOL_IDLE,
     POOL_END,
     POOL_ERROR,
     POOL_FATAL
 } POOL_STATUS;
 
+/* protocol major version numbers */
+#define PROTO_MAJOR_V2	2
+#define PROTO_MAJOR_V3	3
+
+/*
+ * startup packet definitions (v2) stolen from PostgreSQL
+ */
+#define SM_DATABASE		64
+#define SM_USER			32
+#define SM_OPTIONS		64
+#define SM_UNUSED		64
+#define SM_TTY			64
+
+typedef struct StartupPacket_v2
+{
+	int			protoVersion;		/* Protocol version */
+	char		database[SM_DATABASE];	/* Database name */
+	char		user[SM_USER];	/* User name */
+	char		options[SM_OPTIONS];	/* Optional additional args */
+	char		unused[SM_UNUSED];		/* Unused */
+	char		tty[SM_TTY];	/* Tty for debug output */
+} StartupPacket_v2;
+
+/* startup packet info */
+typedef struct
+{
+	char *startup_packet;		/* raw startup packet without packet length (malloced area) */
+	int len;					/* raw startup packet length */
+	int major;	/* protocol major version */
+	int minor;	/* protocol minor version */
+	char *database;	/* database name in startup_packet (malloced area) */
+	char *user;	/* user name in startup_packet (malloced area) */
+} StartupPacket;
+
+typedef struct CancelPacket
+{
+	int			protoVersion;		/* Protocol version */
+	int			pid;	/* bcckend process id */
+	int			key;	/* cancel key */
+} CancelPacket;
+
 /*
  * configuration paramters
  */
@@ -92,32 +133,49 @@ typedef struct {
 	 */
 	int replication_timeout;
 
+	int load_balance_mode;		/* load balance mode */
+
 	/* followings do not exist in the configuration file */
     char	*current_backend_host_name;	/* current backend host name */
     int	current_backend_port;	/* current backend port # */
 	int replication_enabled;		/* replication mode enabled */
+
+	int replication_stop_on_mismatch;		/* if there's a data mismatch between master and secondary
+											 * start degenration to stop replication mode
+											 */
 } POOL_CONFIG;
 
 #define MAX_PASSWORD_SIZE		1024
 
+typedef struct {
+	int num;	/* number of entries */
+	char **names;		/* parameter names */
+	char **values;		/* values */
+} ParamStatus;
+
 /*
  * stream connection structure
  */
 typedef struct {
-	int fd;
-#ifdef NOT_USED
-	FILE *read_fd;	/* stream read connection */
-#endif
+	int fd;		/* fd for connection */
 	FILE *write_fd;	/* stream write connection */
+
 	char *hp;	/* pending data buffer head address */
 	int po;		/* pending data offset */
 	int bufsz;	/* pending data buffer size */
 	int len;	/* pending data length */
+
 	char *sbuf;	/* buffer for pool_read_string */
-	int sbufsz;	/* it's size */
+	int sbufsz;	/* its size in bytes */
+
+	char *buf2;	/* buffer for pool_read2 */
+	int bufsz2;	/* its size in bytes */
+
 	int isbackend;		/* this connection is for backend if non 0 */
 	int issecondary_backend;		/* this connection is for secondary backend if non 0 */
 
+	char tstate;		/* transaction state (V3 only) */
+
 	/*
 	 * following are used to remember when re-use the authenticated connection
 	 */
@@ -126,18 +184,25 @@ typedef struct {
 	char password[MAX_PASSWORD_SIZE];		/* /* password (sent back from frontend) */
 	char salt[4];		/* password salt */
 
+	/*
+	 * following are used to remember current session paramter status.
+	 * re-used connection will need them (V3 only)
+	 */
+	ParamStatus params;
+
+	int no_forward;		/* if non 0, do not write to frontend */
+
 } POOL_CONNECTION;
 
 /*
  * connection pool structure
  */
 typedef struct {
-    char *user;	/* user name */
-    char *database;	/* database name */
+	StartupPacket *sp;	/* startup packet info */
     int pid;	/* backend pid */
     int key;	/* cancel key */
     POOL_CONNECTION	*con;
-	time_t closetime;	/* absolute time in second when the connection closed */
+	time_t closetime;	/* absolute time in seconds when the connection closed */
 } POOL_CONNECTION_POOL_SLOT;
 
 #define MAX_CONNECTION_SLOTS 2
@@ -149,32 +214,11 @@ typedef struct {
 
 #define MASTER_CONNECTION(p) ((p)->slots[0])
 #define SECONDARY_CONNECTION(p) ((p)->slots[1])
-
-/*
- * startup packet definitions stolen from PostgreSQL
- */
-#define SM_DATABASE		64
-#define SM_USER			32
-#define SM_OPTIONS		64
-#define SM_UNUSED		64
-#define SM_TTY			64
-
-typedef struct StartupPacket
-{
-	int			protoVersion;		/* Protocol version */
-	char		database[SM_DATABASE];	/* Database name */
-	char		user[SM_USER];	/* User name */
-	char		options[SM_OPTIONS];	/* Optional additional args */
-	char		unused[SM_UNUSED];		/* Unused */
-	char		tty[SM_TTY];	/* Tty for debug output */
-} StartupPacket;
-
-typedef struct CancelPacket
-{
-	int			protoVersion;		/* Protocol version */
-	int			pid;	/* bcckend process id */
-	int			key;	/* cancel key */
-} CancelPacket;
+#define REPLICATION (pool_config.replication_enabled)
+#define MASTER(p) MASTER_CONNECTION(p)->con
+#define SECONDARY(p) SECONDARY_CONNECTION(p)->con
+#define MAJOR(p) MASTER_CONNECTION(p)->sp->major
+#define TSTATE(p) MASTER(p)->tstate
 
 #define Max(x, y)		((x) > (y) ? (x) : (y))
 #define Min(x, y)		((x) < (y) ? (x) : (y))
@@ -195,26 +239,25 @@ extern int pool_get_config(char *confpath);
 extern void do_child(int unix_fd, int inet_fd);
 extern int pool_init_cp(void);
 extern POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
-				      POOL_CONNECTION_POOL *backend,
-				      int connection_reuse);
+									  POOL_CONNECTION_POOL *backend,
+									  int connection_reuse);
 
 extern POOL_CONNECTION *pool_open(int fd);
 extern void pool_close(POOL_CONNECTION *cp);
 extern int pool_read(POOL_CONNECTION *cp, void *buf, int len);
+extern char *pool_read2(POOL_CONNECTION *cp, int len);
 extern int pool_write(POOL_CONNECTION *cp, void *buf, int len);
 extern int pool_flush(POOL_CONNECTION *cp);
 extern int pool_write_and_flush(POOL_CONNECTION *cp, void *buf, int len);
 extern char *pool_read_string(POOL_CONNECTION *cp, int *len, int line);
 
-extern POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database);
-
-extern POOL_STATUS pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
-extern POOL_STATUS pool_do_reauth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp);
+extern int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
+extern int pool_do_reauth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp);
 
 extern int pool_init_cp(void);
-extern POOL_CONNECTION_POOL *pool_create_cp(char *user, char *database);
-extern POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database);
-extern void pool_discard_cp(char *user, char *database);
+extern POOL_CONNECTION_POOL *pool_create_cp(void);
+extern POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database, int protoMajor);
+extern void pool_discard_cp(char *user, char *database, int protoMajor);
 
 extern POOL_STATUS ErrorResponse(POOL_CONNECTION *frontend, 
 								  POOL_CONNECTION_POOL *backend);
@@ -233,4 +276,26 @@ extern void pool_disable_timeout(void);
 
 extern void pool_send_frontend_exits(POOL_CONNECTION_POOL *backend);
 
+extern int pool_read_message_length(POOL_CONNECTION_POOL *cp);
+extern char pool_read_kind(POOL_CONNECTION_POOL *cp);
+
+extern POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
+extern POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
+extern POOL_STATUS ParameterStatus(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend);
+
+extern int pool_init_params(ParamStatus *params);
+extern void pool_discard_params(ParamStatus *params);
+extern char *pool_find_name(ParamStatus *params, char *name, int *pos);
+extern int pool_get_param(ParamStatus *params, int index, char **name, char **value);
+extern int pool_add_param(ParamStatus *params, char *name, char *value);
+extern void pool_param_debug_print(ParamStatus *params);
+
+extern void pool_send_error_message(POOL_CONNECTION *frontend, int protoMajor,
+							 char *code,
+							 char *message,
+							 char *detail,
+							 char *hint,
+							 char *file,
+							 int line);
+
 #endif /* POOL_H */
