@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/pool_auth.c,v 1.6 2004/05/06 06:33:58 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/pool_auth.c,v 1.11 2004/06/15 12:57:55 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -36,49 +36,40 @@
 #include <errno.h>
 #include <string.h>
 
-static POOL_STATUS pool_send_auth_ok(POOL_CONNECTION *frontend, int pid, int key);
-static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth);
-static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth);
-static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth);
+static POOL_STATUS pool_send_auth_ok(POOL_CONNECTION *frontend, int pid, int key, int protoMajor);
+static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor);
+static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor);
+static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor);
 
 /*
 * do authentication against backend. if success return 0 otherwise non 0.
 */
-POOL_STATUS pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
+int pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 {
 	int status;
 	char kind, kind1;
 	int pid, pid1;
 	int key, key1;
+	int protoMajor;
+	int length;
 
-	status = pool_read(MASTER_CONNECTION(cp)->con, &kind, 1);
-	if (status < 0)
+	protoMajor = MAJOR(cp);
+
+	kind = pool_read_kind(cp);
+	if (kind < 0)
 	{
-		pool_error("pool_do_auth: error while reading \"R\"", strerror(errno));
 		return -1;
 	}
 
-	if (pool_config.replication_enabled)
-	{
-		status = pool_read(SECONDARY_CONNECTION(cp)->con, &kind1, 1);
-		if (status < 0)
-		{
-			pool_error("pool_do_auth: error while reading \"R\" from secondary backend",
-					   strerror(errno));
-			return -1;
-		}
-
-		if (kind != kind1)
-		{
-			pool_error("pool_do_auth: kind does not match between backends master(%c) secondary(%c)",
-					   kind, kind1);
-			return -1;
-		}
-	}
-
-	/* Error response? */
+	/* error response? */
 	if (kind == 'E')
 	{
+		/* we assume error response at this stage is likely version
+		 * protocol mismatch (v3 frontend vs. v2 backend). So we throw
+		 * a V2 protocol error response in the hope that v3 frontend
+		 * will negotiate again using v2 protocol.
+		 */
+		pool_log("pool_do_auth: maybe protocol version mismatch (current version %d)", protoMajor);
 		ErrorResponse(frontend, cp);
 		return -1;
 	}
@@ -88,32 +79,39 @@ POOL_STATUS pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 		return -1;
 	}
 
+	/*
+	 * message length (v3 only) */
+	if (protoMajor == PROTO_MAJOR_V3 && pool_read_message_length(cp) < 0)
+	{
+		return -1;
+	}
+
 	/*
 	 * read authentication request kind.
 	 *
-	 * 0: authenticatoon ok
+	 * 0: authentication ok
 	 * 1: kerberos v4
 	 * 2: kerberos v5
 	 * 3: clear text password
 	 * 4: crypt password
-	 * 5: MD5 password
-	 * 6: SCM credential
+	 * 5: md5 password
+	 * 6: scm credential
 	 *
 	 * in replication mode, we only supports  kind = 0, 3. this is because to "salt"
 	 * cannot be replicated among master and secondary.
 	 * in non replication mode, we supports  kind = 0, 3, 4, 5
 	 */
 
-	status = pool_read(MASTER_CONNECTION(cp)->con, &pid, sizeof(pid));
+	status = pool_read(MASTER(cp), &pid, sizeof(pid));
 	if (status < 0)
 	{
 		pool_error("pool_do_auth: read authentication kind failed");
 		return -1;
 	}
 
-	if (pool_config.replication_enabled)
+	if (REPLICATION)
 	{
-		status = pool_read(SECONDARY_CONNECTION(cp)->con, &pid1, sizeof(pid1));
+		status = pool_read(SECONDARY(cp), &pid1, sizeof(pid1));
 
 		if (status < 0)
 		{
@@ -127,7 +125,20 @@ POOL_STATUS pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	/* trust? */
 	if (pid == 0)
 	{
-		MASTER_CONNECTION(cp)->con->auth_kind = 0;
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			int msglen;
+
+			pool_write(frontend, "R", 1);
+			msglen = htonl(8);
+			pool_write(frontend, &msglen, sizeof(msglen));
+			msglen = htonl(0);
+			if (pool_write_and_flush(frontend, &msglen, sizeof(msglen)) < 0)
+			{
+				return -1;
+			}
+		}
+		MASTER(cp)->auth_kind = 0;
 	}
 
 	/* clear text password authentication? */
@@ -135,11 +146,11 @@ POOL_STATUS pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	{
 		pool_debug("trying clear text password authentication");
 
-		pid = do_clear_text_password(MASTER_CONNECTION(cp)->con, frontend, 0);
+		pid = do_clear_text_password(MASTER(cp), frontend, 0, protoMajor);
 
-		if (pid >= 0 && pool_config.replication_enabled)
+		if (pid >= 0 && REPLICATION)
 		{
-			pid = do_clear_text_password(SECONDARY_CONNECTION(cp)->con, frontend, 0);
+			pid = do_clear_text_password(SECONDARY(cp), frontend, 0, protoMajor);
 		}
 	}
 
@@ -148,153 +159,201 @@ POOL_STATUS pool_do_auth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 	{
 		pool_debug("trying crypt authentication");
 
-		pid = do_crypt(MASTER_CONNECTION(cp)->con, frontend, 0);
+		pid = do_crypt(MASTER(cp), frontend, 0, protoMajor);
 
-		if (pid >= 0 && pool_config.replication_enabled)
+		if (pid >= 0 && REPLICATION)
 		{
-			pid = do_crypt(SECONDARY_CONNECTION(cp)->con, frontend, 0);
+			pid = do_crypt(SECONDARY(cp), frontend, 0, protoMajor);
 		}
 	}
 
-	/* MD5 authentication? */
+	/* md5 authentication? */
 	else if (pid == 5)
 	{
-		pool_debug("trying MD5 authentication");
+		pool_debug("trying md5 authentication");
 
-		pid = do_md5(MASTER_CONNECTION(cp)->con, frontend, 0);
+		pid = do_md5(MASTER(cp), frontend, 0, protoMajor);
 
-		if (pid >= 0 && pool_config.replication_enabled)
+		if (pid >= 0 && REPLICATION)
 		{
-			pid = do_md5(SECONDARY_CONNECTION(cp)->con, frontend, 0);
+			pid = do_md5(SECONDARY(cp), frontend, 0, protoMajor);
 		}
 	}
 
 	if (pid != 0)
 	{
-		pool_error("pool_do_auth: backend does not return Authenticaton Ok");
+		pool_error("pool_do_auth: backend does not return authenticaton ok");
 		return -1;
 	} 
 
 	/*
-	 * Authentication OK. Now read pid and secret key from the
+	 * authentication ok. now read pid and secret key from the
 	 * backend
 	 */
-	status = pool_read(MASTER_CONNECTION(cp)->con, &kind, 1);
-	if (status < 0)
+	kind = pool_read_kind(cp);
+	if (kind < 0)
 	{
-		pool_error("pool_do_auth: error while reading \"W\"");
 		return -1;
 	}
-	if (pool_config.replication_enabled)
-	{
-		status = pool_read(SECONDARY_CONNECTION(cp)->con, &kind1, 1);
-		if (status < 0)
-		{
-			pool_error("pool_do_auth: error while reading \"W\" from secondary backend",
-					   strerror(errno));
-			return -1;
-		}
 
-		if (kind != kind1)
-		{
-			pool_error("pool_do_auth: kind is not match between backends master(%c) secondary(%c)",
-					   kind, kind1);
-			return -1;
-		}
-	}
-
-	/* Error response? */
+	/* error response? */
 	if (kind == 'E')
 	{
-		ErrorResponse(frontend, cp);
+		if (protoMajor == PROTO_MAJOR_V2)
+			ErrorResponse(frontend, cp);
+		else
+			SimpleForwardToFrontend(kind, frontend, cp);
 		return -1;
 	}
 	else if (kind != 'K')
 	{
-		pool_error("pool_do_auth: expect \"K\" got %c", kind);
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			char *p;
+			int len;
+
+			/* process parameter status */
+			while (kind == 'S')
+			{
+				if (ParameterStatus(frontend, cp) != POOL_CONTINUE)
+						return -1;
+
+				pool_flush(frontend);
+
+				kind = pool_read_kind(cp);
+				if (kind < 0)
+				{
+					pool_error("pool_do_auth: failed to read kind while processing ParamterStatus");
+					return -1;
+				}
+			}
+		}
+		else
+		{
+			pool_error("pool_do_auth: expect \"K\" got %c", kind);
+			return -1;
+		}
+	}
+
+	/*
+	 * message length (V3 only)
+	 */
+	if (protoMajor == PROTO_MAJOR_V3 && (length = pool_read_message_length(cp)) != 12)
+	{
+		pool_error("pool_do_auth: invalid messages length(%d) for BackendKeyData", length);
 		return -1;
 	}
 
 	/*
 	 * OK, read pid and secret key
 	 */
+
 	/* pid */
-	pool_read(MASTER_CONNECTION(cp)->con, &pid, sizeof(pid));
+	pool_read(MASTER(cp), &pid, sizeof(pid));
 	MASTER_CONNECTION(cp)->pid = pid;
 
 	/* key */
-	pool_read(MASTER_CONNECTION(cp)->con, &key, sizeof(key));
+	pool_read(MASTER(cp), &key, sizeof(key));
 	MASTER_CONNECTION(cp)->key = key;
 
-	if (pool_config.replication_enabled)
+	if (REPLICATION)
 	{
-		pool_read(SECONDARY_CONNECTION(cp)->con, &pid1, sizeof(pid1));
+		pool_read(SECONDARY(cp), &pid1, sizeof(pid1));
 		SECONDARY_CONNECTION(cp)->pid = pid;
 
 		/* key */
-		pool_read(SECONDARY_CONNECTION(cp)->con, &key1, sizeof(key1));
+		pool_read(SECONDARY(cp), &key1, sizeof(key1));
 		SECONDARY_CONNECTION(cp)->key = key;
 	}
 
-	return (pool_send_auth_ok(frontend, pid, key));
+	return (pool_send_auth_ok(frontend, pid, key, protoMajor));
 }
 
 /*
 * do re-authentication for reused connection. if success return 0 otherwise non 0.
 */
-POOL_STATUS pool_do_reauth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
+int pool_do_reauth(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *cp)
 {
 	int status;
+	int protoMajor;
+
+	protoMajor = MAJOR(cp);
 
 	/* trust? */
-	if (MASTER_CONNECTION(cp)->con->auth_kind == 0)
+	if (MASTER(cp)->auth_kind == 0)
 		status = 0;
 
 	/* clear text password? */
-	else if (MASTER_CONNECTION(cp)->con->auth_kind == 3)
+	else if (MASTER(cp)->auth_kind == 3)
 	{
-		status = do_clear_text_password(MASTER_CONNECTION(cp)->con, frontend, 1);
+		status = do_clear_text_password(MASTER(cp), frontend, 1, protoMajor);
 	}
 
 	/* crypt password? */
-	else if (MASTER_CONNECTION(cp)->con->auth_kind == 4)
+	else if (MASTER(cp)->auth_kind == 4)
 	{
-		status = do_crypt(MASTER_CONNECTION(cp)->con, frontend, 1);
+		status = do_crypt(MASTER(cp), frontend, 1, protoMajor);
 	}
 
 	/* md5 password? */
-	else if (MASTER_CONNECTION(cp)->con->auth_kind == 5)
+	else if (MASTER(cp)->auth_kind == 5)
 	{
-		status = do_md5(MASTER_CONNECTION(cp)->con, frontend, 1);
+		status = do_md5(MASTER(cp), frontend, 1, protoMajor);
 	}
 
 	if (status == 0)
-		return (pool_send_auth_ok(frontend, MASTER_CONNECTION(cp)->pid, MASTER_CONNECTION(cp)->key) != POOL_CONTINUE);
-	return -1;
+	{
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			int msglen;
+
+			pool_write(frontend, "R", 1);
+			msglen = htonl(8);
+			pool_write(frontend, &msglen, sizeof(msglen));
+			msglen = htonl(0);
+			if (pool_write_and_flush(frontend, &msglen, sizeof(msglen)) < 0)
+			{
+				return -1;
+			}
+		}
+	}
+	else
+	{
+		pool_debug("pool_do_reauth: authentication failed");
+		return -1;
+	}
+
+	return (pool_send_auth_ok(frontend, MASTER_CONNECTION(cp)->pid, MASTER_CONNECTION(cp)->key, protoMajor) != POOL_CONTINUE);
 }
 
 /*
 * send authentication ok to frontend. if success return 0 otherwise non 0.
 */
-static POOL_STATUS pool_send_auth_ok(POOL_CONNECTION *frontend, int pid, int key)
+static POOL_STATUS pool_send_auth_ok(POOL_CONNECTION *frontend, int pid, int key, int protoMajor)
 {
 	char kind;
-	char buff[128];
 	int len;
 
-	/* return "Authentication OK" to the frontend */
-	kind = 'R';
-	pool_write(frontend, &kind, 1);
-	len = htonl(0);
-	memcpy(buff, &len, sizeof(len));
-	if (pool_write_and_flush(frontend, buff, sizeof(len)) < 0)
+	if (protoMajor == PROTO_MAJOR_V2)
 	{
-		return -1;
+		/* return "Authentication OK" to the frontend */
+		kind = 'R';
+		pool_write(frontend, &kind, 1);
+		len = htonl(0);
+		if (pool_write_and_flush(frontend, &len, sizeof(len)) < 0)
+		{
+			return -1;
+		}
 	}
 
-	/* Backend Key Data */
+	/* send backend key data */
 	kind = 'K';
 	pool_write(frontend, &kind, 1);
+	if (protoMajor == PROTO_MAJOR_V3)
+	{
+		len = htonl(12);
+		pool_write(frontend, &len, sizeof(len));
+	}
 	pool_write(frontend, &pid, sizeof(pid));
 	if (pool_write_and_flush(frontend, &key, sizeof(key)) < 0)
 	{
@@ -307,30 +366,59 @@ static POOL_STATUS pool_send_auth_ok(POOL_CONNECTION *frontend, int pid, int key
 /*
  * perform clear text password authetication
  */
-static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth)
+static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
 	static int size;
 	static char password[MAX_PASSWORD_SIZE];
 	char response;
 	int kind;
+	int len;
 
 	/* master? */
 	if (!backend->issecondary_backend)
 	{
 		pool_write(frontend, "R", 1);	/* authenticaton */
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			len = htonl(8);
+			pool_write(frontend, &len, sizeof(len));
+		}
 		kind = htonl(3);		/* clear text password authentication */
 		pool_write_and_flush(frontend, &kind, sizeof(kind));	/* indicating clear text password authentication */
 
 		/* read password packet */
-		if (pool_read(frontend, &size, sizeof(size)))
+		if (protoMajor == PROTO_MAJOR_V2)
 		{
-			pool_error("do_clear_text_password: failed to read password packet size");
-			return -1;
+			if (pool_read(frontend, &size, sizeof(size)))
+			{
+				pool_error("do_clear_text_password: failed to read password packet size");
+				return -1;
+			}
+		}
+		else
+		{
+			char k;
+
+			if (pool_read(frontend, &k, sizeof(k)))
+			{
+				pool_error("do_clear_text_password: failed to read password packet \"p\"");
+				return -1;
+			}
+			if (k != 'p')
+			{
+				pool_error("do_clear_text_password: password packet does not start with \"p\"");
+				return -1;
+			}
+			if (pool_read(frontend, &size, sizeof(size)))
+			{
+				pool_error("do_clear_text_password: failed to read password packet size");
+				return -1;
+			}
 		}
 
 		if ((ntohl(size) - 4) > sizeof(password))
 		{
-			pool_error("do_clear_text_password: password is too long(size: %d)", ntohl(size) - 4);
+			pool_error("do_clear_text_password: password is too long (size: %d)", ntohl(size) - 4);
 			return -1;
 		}
 
@@ -360,6 +448,8 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 	}
 
 	/* send password packet to backend */
+	if (protoMajor == PROTO_MAJOR_V3)
+		pool_write(backend, "p", 1);
 	pool_write(backend, &size, sizeof(size));
 	pool_write_and_flush(backend, password, ntohl(size) -4);
 	if (pool_read(backend, &response, sizeof(response)))
@@ -374,6 +464,21 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 		return -1;
 	}
 
+	if (protoMajor == PROTO_MAJOR_V3)
+	{
+		if (pool_read(backend, &len, sizeof(len)))
+		{
+			pool_error("do_clear_text_password: failed to read authentication packet size");
+			return -1;
+		}
+
+		if (ntohl(len) != 8)
+		{
+			pool_error("do_clear_text_password: incorrect authentication packet size (%d)", ntohl(len));
+			return -1;
+		}
+	}
+
 	/* expect to read "Authentication OK" response. kind should be 0... */
 	if (pool_read(backend, &kind, sizeof(kind)))
 	{
@@ -384,6 +489,20 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 	/* if authenticated, save info */
 	if (!reauth && kind == 0)
 	{
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			int msglen;
+
+			pool_write(frontend, "R", 1);
+			msglen = htonl(8);
+			pool_write(frontend, &msglen, sizeof(msglen));
+			msglen = htonl(0);
+			if (pool_write_and_flush(frontend, &msglen, sizeof(msglen)) < 0)
+			{
+				return -1;
+			}
+		}
+
 		backend->auth_kind = 3;
 		backend->pwd_size = ntohl(size) - 4;
 		memcpy(backend->password, password, backend->pwd_size);
@@ -394,13 +513,14 @@ static int do_clear_text_password(POOL_CONNECTION *backend, POOL_CONNECTION *fro
 /*
  * perform crypt authetication
  */
-static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth)
+static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
 	char salt[2];
 	static int size;
 	static char password[MAX_PASSWORD_SIZE];
 	char response;
 	int kind;
+	int len;
 
 	if (!reauth)
 	{
@@ -420,15 +540,43 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 	if (!backend->issecondary_backend)
 	{
 		pool_write(frontend, "R", 1);	/* authenticaton */
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			len = htonl(10);
+			pool_write(frontend, &len, sizeof(len));
+		}
 		kind = htonl(4);		/* crypt authentication */
 		pool_write(frontend, &kind, sizeof(kind));	/* indicating crypt authentication */
 		pool_write_and_flush(frontend, salt, sizeof(salt));		/* salt */
 
 		/* read password packet */
-		if (pool_read(frontend, &size, sizeof(size)))
+		if (protoMajor == PROTO_MAJOR_V2)
 		{
-			pool_error("do_crypt: failed to read password packet size");
-			return -1;
+			if (pool_read(frontend, &size, sizeof(size)))
+			{
+				pool_error("do_crypt: failed to read password packet size");
+				return -1;
+			}
+		}
+		else
+		{
+			char k;
+
+			if (pool_read(frontend, &k, sizeof(k)))
+			{
+				pool_error("do_crypt_password: failed to read password packet \"p\"");
+				return -1;
+			}
+			if (k != 'p')
+			{
+				pool_error("do_crypt_password: password packet does not start with \"p\"");
+				return -1;
+			}
+			if (pool_read(frontend, &size, sizeof(size)))
+			{
+				pool_error("do_crypt_password: failed to read password packet size");
+				return -1;
+			}
 		}
 
 		if ((ntohl(size) - 4) > sizeof(password))
@@ -464,6 +612,8 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 	}
 
 	/* send password packet to backend */
+	if (protoMajor == PROTO_MAJOR_V3)
+		pool_write(backend, "p", 1);
 	pool_write(backend, &size, sizeof(size));
 	pool_write_and_flush(backend, password, ntohl(size) -4);
 	if (pool_read(backend, &response, sizeof(response)))
@@ -478,6 +628,21 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 		return -1;
 	}
 
+	if (protoMajor == PROTO_MAJOR_V3)
+	{
+		if (pool_read(backend, &len, sizeof(len)))
+		{
+			pool_error("do_clear_text_password: failed to read authentication packet size");
+			return -1;
+		}
+
+		if (ntohl(len) != 8)
+		{
+			pool_error("do_clear_text_password: incorrect authentication packet size (%d)", ntohl(len));
+			return -1;
+		}
+	}
+
 	/* expect to read "Authentication OK" response. kind should be 0... */
 	if (pool_read(backend, &kind, sizeof(kind)))
 	{
@@ -486,8 +651,21 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 	}
 
 	/* if authenticated, save info */
-	if (kind == 0)
+	if (!reauth && kind == 0)
 	{
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			int msglen;
+
+			pool_write(frontend, "R", 1);
+			msglen = htonl(8);
+			pool_write(frontend, &msglen, sizeof(msglen));
+			msglen = htonl(0);
+			if (pool_write_and_flush(frontend, &msglen, sizeof(msglen)) < 0)
+			{
+				return -1;
+			}
+		}
 		backend->auth_kind = 4;
 		backend->pwd_size = ntohl(size) - 4;
 		memcpy(backend->password, password, backend->pwd_size);
@@ -499,19 +677,23 @@ static int do_crypt(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int rea
 /*
  * perform MD5 authetication
  */
-static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth)
+static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reauth, int protoMajor)
 {
 	char salt[4];
 	static int size;
 	static char password[MAX_PASSWORD_SIZE];
 	char response;
 	int kind;
+	int len;
 
-	/* read salt */
-	if (pool_read(backend, salt, sizeof(salt)))
+	if (!reauth)
 	{
-		pool_error("do_md5: failed to read salt");
-		return -1;
+		/* read salt */
+		if (pool_read(backend, salt, sizeof(salt)))
+		{
+			pool_error("do_md5: failed to read salt");
+			return -1;
+		}
 	}
 	else
 	{
@@ -522,15 +704,43 @@ static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reaut
 	if (!backend->issecondary_backend)
 	{
 		pool_write(frontend, "R", 1);	/* authenticaton */
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			len = htonl(12);
+			pool_write(frontend, &len, sizeof(len));
+		}
 		kind = htonl(5);
 		pool_write(frontend, &kind, sizeof(kind));	/* indicating MD5 */
 		pool_write_and_flush(frontend, salt, sizeof(salt));		/* salt */
 
 		/* read password packet */
-		if (pool_read(frontend, &size, sizeof(size)))
+		if (protoMajor == PROTO_MAJOR_V2)
 		{
-			pool_error("do_md5: failed to read password packet size");
-			return -1;
+			if (pool_read(frontend, &size, sizeof(size)))
+			{
+				pool_error("do_md5: failed to read password packet size");
+				return -1;
+			}
+		}
+		else
+		{
+			char k;
+
+			if (pool_read(frontend, &k, sizeof(k)))
+			{
+				pool_error("do_md5_password: failed to read password packet \"p\"");
+				return -1;
+			}
+			if (k != 'p')
+			{
+				pool_error("do_md5_password: password packet does not start with \"p\"");
+				return -1;
+			}
+			if (pool_read(frontend, &size, sizeof(size)))
+			{
+				pool_error("do_md5_password: failed to read password packet size");
+				return -1;
+			}
 		}
 
 		if ((ntohl(size) - 4) > sizeof(password))
@@ -565,6 +775,8 @@ static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reaut
 	}
 
 	/* send password packet to backend */
+	if (protoMajor == PROTO_MAJOR_V3)
+		pool_write(backend, "p", 1);
 	pool_write(backend, &size, sizeof(size));
 	pool_write_and_flush(backend, password, ntohl(size) -4);
 	if (pool_read(backend, &response, sizeof(response)))
@@ -575,10 +787,25 @@ static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reaut
 
 	if (response != 'R')
 	{
-		pool_debug("do_md5: backend does not return R while processing MD5 authentication");
+		pool_debug("do_md5: backend does not return R while processing MD5 authentication %c", response);
 		return -1;
 	}
 
+	if (protoMajor == PROTO_MAJOR_V3)
+	{
+		if (pool_read(backend, &len, sizeof(len)))
+		{
+			pool_error("do_md5: failed to read authentication packet size");
+			return -1;
+		}
+
+		if (ntohl(len) != 8)
+		{
+			pool_error("do_clear_text_password: incorrect authentication packet size (%d)", ntohl(len));
+			return -1;
+		}
+	}
+
 	/* expect to read "Authentication OK" response. kind should be 0... */
 	if (pool_read(backend, &kind, sizeof(kind)))
 	{
@@ -587,8 +814,21 @@ static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reaut
 	}
 
 	/* if authenticated, save info */
-	if (kind == 0)
+	if (!reauth && kind == 0)
 	{
+		if (protoMajor == PROTO_MAJOR_V3)
+		{
+			int msglen;
+
+			pool_write(frontend, "R", 1);
+			msglen = htonl(8);
+			pool_write(frontend, &msglen, sizeof(msglen));
+			msglen = htonl(0);
+			if (pool_write_and_flush(frontend, &msglen, sizeof(msglen)) < 0)
+			{
+				return -1;
+			}
+		}
 		backend->auth_kind = 5;
 		backend->pwd_size = ntohl(size) - 4;
 		memcpy(backend->password, password, backend->pwd_size);
@@ -596,3 +836,78 @@ static int do_md5(POOL_CONNECTION *backend, POOL_CONNECTION *frontend, int reaut
 	}
 	return kind;
 }
+
+/*
+ * read message length (V3 only)
+ */
+int pool_read_message_length(POOL_CONNECTION_POOL *cp)
+{
+	int status;
+	int length, length1;
+
+	status = pool_read(MASTER(cp), &length, sizeof(length));
+	if (status < 0)
+	{
+		pool_error("read_message_length: error while reading message length");
+		return -1;
+	}
+	length = ntohl(length);
+
+	if (REPLICATION)
+	{
+		status = pool_read(SECONDARY(cp), &length1, sizeof(length1));
+		if (status < 0)
+		{
+			pool_error("read_message_length: error while reading message length from secondary backend");
+			return -1;
+		}
+		length1 = ntohl(length1);
+
+		if (length != length1)
+		{
+			pool_error("read_message_length: length does not match between backends master(%d) secondary(%d)",
+					   length, length1);
+			return -1;
+		}
+	}
+
+	if (length < 0)
+	{
+		pool_error("read_message_length: invalid message length (%d)", length);
+		return -1;
+	}
+
+	return length;
+}
+
+char pool_read_kind(POOL_CONNECTION_POOL *cp)
+{
+	int status;
+	char kind, kind1;
+
+	status = pool_read(MASTER(cp), &kind, sizeof(kind));
+	if (status < 0)
+	{
+		pool_error("read_message_kind: error while reading message kind");
+		return -1;
+	}
+
+	if (REPLICATION)
+	{
+		status = pool_read(SECONDARY(cp), &kind1, sizeof(kind1));
+		if (status < 0)
+		{
+			pool_error("read_message_kind: error while reading message kind from secondary backend");
+			return -1;
+		}
+
+		if (kind != kind1)
+		{
+			pool_error("read_message_kind: kind does not match between backends master(%d) secondary(%d)",
+					   kind, kind1);
+			return -1;
+		}
+	}
+
+	return kind;
+}
