@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/pool_process_query.c,v 1.22 2004/06/18 12:11:06 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/pool_process_query.c,v 1.23 2004/07/02 12:50:09 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -1726,21 +1726,38 @@ void pool_send_frontend_exits(POOL_CONNECTION_POOL *backend)
  */
 POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend)
 {
-	int len;
-	int sendlen;
+	int len, len1;
 	char *p;
+	int status;
 
 	pool_write(frontend, &kind, 1);
 
-	len = pool_read_message_length(backend);
-	if (len < 0)
+	status = pool_read(MASTER(backend), &len, sizeof(len));
+	if (status < 0)
 	{
+		pool_error("SimpleForwardToFrontend: error while reading message length");
 		return POOL_END;
 	}
 
-	sendlen = htonl(len);
-	pool_write(frontend, &sendlen, sizeof(sendlen));
+	if (REPLICATION)
+	{
+		status = pool_read(SECONDARY(backend), &len1, sizeof(len1));
+		if (status < 0)
+		{
+			pool_error("SimpleForwardToFrontend: error while reading message length from secondary backend");
+			return POOL_END;
+		}
 
+		if (len != len1)
+		{
+			pool_log("SimpleForwardToFrontend: length does not match between backends master(%d) secondary(%d)",
+					   ntohl(len), ntohl(len1));
+		}
+	}
+
+	pool_write(frontend, &len, sizeof(len));
+
+	len = ntohl(len);
 	len -= 4;
 
 	p = pool_read2(MASTER(backend), len);
@@ -1748,8 +1765,12 @@ POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_C
 		return POOL_END;
 
 	if (REPLICATION)
-		if (pool_read2(SECONDARY(backend), len) == NULL)
+	{
+		len1 = ntohl(len1);
+		len1 -= 4;
+		if (pool_read2(SECONDARY(backend), len1) == NULL)
 			return POOL_END;
+	}
 
 	return pool_write(frontend, p, len);
 }
